import logging
import os
from pathlib import Path
from typing import Dict, TypeVar

from confme import BaseConfig

T = TypeVar('T', bound='BaseConfig')


class GlobalConfig(BaseConfig):
    _KEY_LOOKUP = ['env', 'environment', 'environ', 'stage']
    _config_path: Path = None
    _cache: Dict[str, T] = {}

    @classmethod
    def register_folder(cls, config_folder: Path):
        cls._config_path = config_folder
        cls._cache = {}

    @classmethod
    def get(cls) -> T:
        env = cls._get_current_env()
        if env not in cls._cache:
            cls._cache[env] = cls._load_file(env)

        return cls._cache[env]

    @classmethod
    def _load_file(cls, environment: str) -> T:
        files = Path(cls._config_path).glob(pattern='*')
        selected_files = [f for f in files if environment in f.name]

        if len(selected_files) <= 0:
            raise Exception(f'No configuration found for environment {environment} in '
                            f'files {files}')
        elif len(selected_files) > 1:
            logging.warning(f'More than one file found matching environment {environment} in'
                            f'files {files}. Using file {selected_files[0]}')
            file = selected_files[0]
        else:
            file = selected_files[0]

        return cls.load(file)

    @classmethod
    def _get_current_env(cls) -> str:
        keys = [key for key in os.environ.keys() if key.lower() in cls._KEY_LOOKUP]

        if len(keys) <= 0:
            raise Exception(f"You're using the register_folder / get combination one of the "
                            f"following environment variables need to be set: ENV, ENVIRONMENT,"
                            f"ENVIRON, env, environment, environ")
        elif len(keys) > 1:
            logging.warning(f'More than one environment variable set using the value of {keys[0]}')
            key = keys[0]
        else:
            key = keys[0]

        return os.environ[key].lower()



from typing import Dict

from pydantic.main import ModelMetaclass
from pydantic.schema import model_schema


def _create_dict(schema_head: Dict, definitions: Dict):
    schema_dict = {}
    if 'type' in schema_head and schema_head['type'] == 'object':
        for key, value in schema_head['properties'].items():
            if 'allOf' in value:
                sub_schema = value['allOf'][0]['$ref'].replace('#/definitions/', '')
                schema_dict[key] = _create_dict(definitions[sub_schema], definitions)
            else:
                schema_dict[key] = None
    else:
        return None
    return schema_dict


def get_schema(config_cls: ModelMetaclass):
    schema = model_schema(config_cls)
    definitions = schema['definitions'] if 'definitions' in schema else {}
    return _create_dict(schema, definitions)



# module of all supported annotations

import pydantic
from pydantic import Field


def Secret(env_var: str):
    return Field(..., env=env_var)


def OpenRange(gt: float = None, lt: float = None) -> Field:
    return Field(..., gt=gt, lt=lt)


def ClosedRange(ge: float = None, le: float = None):
    return Field(..., ge=ge, le=le)


def MixedRange(gt: float = None, ge: float = None, lt: float = None, le: float = None):
    return Field(..., gt=gt, ge=ge, lt=lt, le=le)


__all__ = pydantic.__all__ + [Secret, OpenRange, ClosedRange, MixedRange]



"""base module for all file backends e.g. yaml, json, xml, ..."""
from abc import abstractmethod
from io import StringIO
from typing import Dict, List


class BaseFileParser:
    """Base class for all file backends"""

    @abstractmethod
    def get_endings(self) -> List[str]:
        """Returns a list of file endings (including point e.g. .json)
        supported by the given backend
        :return: List of supported file endings
        """
        pass

    @abstractmethod
    def parse(self, file: StringIO) -> Dict:
        """Base method for converting the given string into a python dict
        :param file: StringIO with access to the raw file data
        :return: Dict of file content
        """
        pass



import argparse

from pydantic.main import ModelMetaclass

from confme.utils.dict_util import flatten, InfiniteDict
from confme.utils.typing import get_schema


def argument_overwrite(config_cls: ModelMetaclass):
    # extract possible parameters
    config_dict = get_schema(config_cls)
    parameters, _ = flatten(config_dict)

    # get arguments from command line
    parser = argparse.ArgumentParser(prefix_chars='+/')
    group = parser.add_argument_group('Configuration Parameters',
                                      'With the parameters specified bellow, '
                                      'the configuration values from the config file can be overwritten.')
    for param in parameters:
        group.add_argument(f'++{param}', required=False)
    args, unknown = parser.parse_known_args()

    # find passed arguments and fill it into the dict structure
    infinite_dict = InfiniteDict()
    for param in parameters:
        value = getattr(args, param)
        if value:
            infinite_dict.expand(param.split('.'), value)

    return infinite_dict



import logging
import os
from pathlib import Path
from typing import Union, Any, List, Tuple, Callable, TypeVar, Dict

from pydantic import BaseSettings
from tabulate import tabulate

from confme import source_backend
from confme.core.argument_overwrite import argument_overwrite
from confme.core.env_overwrite import env_overwrite
from confme.utils.base_exception import ConfmeException
from confme.utils.dict_util import recursive_update, flatten

T = TypeVar('T', bound='BaseConfig')


class BaseConfig(BaseSettings):
    _KEY_LOOKUP = ['env', 'environment', 'environ', 'stage']
    _config_path: Path = None
    _default_env: str = None
    _cache: Dict[str, T] = {}

    @classmethod
    def load(cls, path: Union[Path, str]) -> 'BaseConfig':
        """Load your configuration file into your config class structure.
        :param config_class: Root class to map the configuration file to
        :param path: path to configuration file
        :return: instance of config_class with all values added from the config file
        """
        config_content = source_backend.parse_file(Path(path))
        config_content = recursive_update(config_content, env_overwrite(cls))
        config_content = recursive_update(config_content, argument_overwrite(cls))

        return cls.parse_obj(config_content)

    @classmethod
    def load_from_dict(cls, config_content: Dict) -> 'BaseConfig':
        config_content = recursive_update(config_content, env_overwrite(cls))
        config_content = recursive_update(config_content, argument_overwrite(cls))

        return cls.parse_obj(config_content)

    @classmethod
    def register_folder(cls, config_folder: Path, default_env: str = None):
        """Register a folder where configuration files are drawn based on the environment.
        :param config_folder: Path to the folder with configuration files per environment
        :param default_env: Default environment that should be used if none is specified via environment variable
        """
        cls._config_path = config_folder
        cls._default_env = default_env
        cls._cache = {}

    @classmethod
    def get(cls) -> T:
        """Get the corresponding configuration based on the environment. Thereby, the configuration class is loaded
        once and cached for subsequent calls.
        :return: instance of config_class with all values added from the config file
        """
        env = cls._get_current_env()
        if env not in cls._cache:
            cls._cache[env] = cls._load_file(env)

        return cls._cache[env]

    @classmethod
    def _load_file(cls, environment: str) -> T:
        files = Path(cls._config_path).glob(pattern='*')
        selected_files = [f for f in files if environment in f.name]

        if len(selected_files) <= 0:
            raise Exception(f'No configuration found for environment {environment} in '
                            f'files {files}')
        elif len(selected_files) > 1:
            logging.warning(f'More than one file found matching environment {environment} in'
                            f'files {files}. Using file {selected_files[0]}')
            file = selected_files[0]
        else:
            file = selected_files[0]

        return cls.load(file)

    @classmethod
    def _get_current_env(cls) -> str:
        keys = [key for key in os.environ.keys() if key.lower() in cls._KEY_LOOKUP]

        if len(keys) <= 0 and cls._default_env is None:
            raise Exception(f"You're using the register_folder / get combination one of the "
                            f"following environment variables need to be set: ENV, ENVIRONMENT,"
                            f"ENVIRON, env, environment, environ")
        elif len(keys) <= 0 and cls._default_env is not None:
            return cls._default_env
        elif len(keys) > 1:
            logging.warning(f'More than one environment variable set using the value of {keys[0]}')
            key = keys[0]
        else:
            key = keys[0]

        return os.environ[key].lower()

    def update_by_str(self, path: str, value: Any):
        """Given a path string separated by dots (.) this method allows to update nested configuration objects.
        CAVEAT: For this update no type check is applied!
        e.g. given this configuration
        ```
        class DatabaseConfig(BaseConfig):
            host: str
            port: int
            user: str

        class MyConfig(BaseConfig):
            name: int
            database: DatabaseConfig

        config = MyConfig.load('test.yaml')

        config.update_by_str('database.host', 'my new host')
        :param path: dot (.) separated string which value should be updated
        :param value: update value
        """
        path = path.split('.')
        current = self
        for i, segment in enumerate(path):
            if not hasattr(current, segment):
                raise ConfmeException(f'{segment} not found in path {".".join(path[:i])}!')

            # for the last item we don't want to assign it to the current element because we want to assign the given
            # value instead
            if i + 1 < len(path):
                current = getattr(current, segment)
            else:
                setattr(current, segment, value)

    def get_flat_repr(self) -> List[Tuple[str, Any]]:
        """Returns a flat representation of your configuration structure (tree).
        e.g. given this configuration
        ```
        class DatabaseConfig(BaseConfig):
            host: str
            port: int
            user: str

        class MyConfig(BaseConfig):
            name: int
            database: DatabaseConfig

        config = MyConfig.load('test.yaml')
        config.get_flat_repr()
        ...
        [('name', 'test1'),
        ('database.host', 'localhost'),
        ('database.port', '5678'),
        ('database.user', 'db_user')]

        :return: list of configuration parameter key and it's corresponding value
        """
        keys, values = flatten(self.dict())
        return list(zip(keys, values))

    def log_config(self, print_fn: Callable[[str], None] = logging.info):
        """Prints/logs the configuration in a flat format.
        :param print_fn: print callable to overwrite can be used e.g. with log_config(print_fn=print)
        """
        flat_config = self.get_flat_repr()
        str_config = tabulate(flat_config, headers=['Key', 'Value'], tablefmt="github")
        print_fn(str_config)



"""confme package exports"""
from confme.core.base_config import BaseConfig
from confme.core.global_config import GlobalConfig
from confme.utils.base_exception import ConfmeException

__all__ = [BaseConfig, GlobalConfig, ConfmeException]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['confme', 'confme.core', 'confme.source_backend', 'confme.utils']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.4,<2.0', 'pyyaml>=5.3,<6.0', 'tabulate>=0.8.9,<0.9.0']

setup_kwargs = {
    'name': 'confme',
    'version': '1.3.1',
    'description': 'Easy configuration management in python',
    'long_description': '# ConfMe: Configuration Made Easy 💖\n\n[![image](https://img.shields.io/pypi/v/confme?color=blue)](https://pypi.org/project/confme/) [![image](https://img.shields.io/pypi/l/confme)](https://pypi.org/project/confme/) [![image](https://github.com/iwanbolzern/ConfMe/workflows/Test/badge.svg?branch=master)](https://pypi.org/project/confme/)\n[![image](https://img.shields.io/pypi/pyversions/confme?color=blue)](https://pypi.org/project/confme/)\n\nConfMe is a simple to use, production ready application configuration management library, which takes into consideration the following three thoughts:\n1. Access to configuration values must be safe at runtime. **No ```myconfig[\'value1\'][\'subvalue\']``` anymore!**\n2. The configuration must be checked for consistency at startup e.g. type check, range check, ...\n3. Secrets shall be injectable from environment variables\n\nConfMe makes all these features possible with just a few type annotations on plain Python objects.\n\n## Installation\nConfMe can be installed from the official python package repository [pypi](https://pypi.org/project/confme/)\n\n```\npip install confme\n```\n\nOr, if you\'re using pipenv:\n\n```\npipenv install confme\n```\n\nOr, if you\'re using poetry:\n\n```\npoetry add confme\n```\n\n## Basic Usage of confme\nDefine your config structure as plain python objects with type annotations:\n\n```python\nfrom confme import BaseConfig\n\nclass DatabaseConfig(BaseConfig):\n    host: str\n    port: int\n    user: str\n\nclass MyConfig(BaseConfig):\n    name: str\n    database: DatabaseConfig\n```\n\nCreate a configuration yaml file with the same structure as your configuration classes have:\n\n```yaml\nname: "Database Application"\ndatabase:\n    host: "localhost"\n    port: 5000\n    user: "any-db-user"\n```\n\nLoad the yaml file into your Python object structure and access it in a secure manner:\n\n```python\nmy_config = MyConfig.load(\'config.yaml\')\n\nprint(f\'Using database connection {my_config.database.host} \'\n      f\'on port {my_config.database.port}\')\n```\n\nIn the background the yaml file is parsed and mapped to the defined object structure. While mapping the values to object properties, type checks are performed. If a value is not available or is not of the correct type, an error is generated already when the configuration is loaded.\n\n## Supported Annotations\nConfMe is based on pydantic and supports all annotations provided by pydantic. The most important annotations are listed and explain bellow. For the whole list, please checkout [Field Types](https://pydantic-docs.helpmanual.io/usage/types/):\n- str\n- int\n- float\n- bool\n- typing.List[x]\n- typing.Optional[x]\n- [Secret](#secret)\n- [Range](#range)\n- [Enum](#enum)\n\n### Secret\nWith the Secret annotation you can inject secrets from environment variables directly into your configuration structure. This is especially handy when you\'re deploying applications by using docker. Therefore, let\'s extend the previous example with a Secret annotation:\n\n```python\nfrom confme import BaseConfig\nfrom confme.annotation import Secret\n\nclass DatabaseConfig(BaseConfig):\n    ...\n    password: str = Secret(\'highSecurePassword\')\n```\n\nNow set the password to the defined environment variable:\n\n```bash\nexport highSecurePassword="This is my password"\n```\n\nLoad your config and check for the injected password.\n\n```\nmy_config = MyConfig.load(\'config.yaml\')\nprint(f\'My password is: {my_config.database.password}\')\n```\n\n### Range\n\nConfME supports OpenRange, ClosedRange and MixedRange values. The terms open and close are similar to open and closed intervals in mathematics. This means, if you want to include the lower and upper range use ClosedRange otherwise OpenRange:\n* ```ClosedRange(2, 3)``` will include 2 and 3\n* ```OpenRange(2, 3)``` will not include 2 and 3\n\nIf you want to have a mixture of both, e.g. include 2 but exclude 3 use MixedRange:\n* ```MixedRange(ge=2, lt=3)``` will include 2 but exclude 3\n\n```python\nfrom confme import BaseConfig\nfrom confme.annotation import ClosedRange\n\nclass DatabaseConfig(BaseConfig):\n    ...\n    password: int = ClosedRange(2, 3)\n```\n\n### Enum\n\nIf a Python Enum is set as type annotation, ConfMe expect to find the enum value in the configuration file.\n\n```python\nfrom confme import BaseConfig\nfrom enum import Enum\n\nclass DatabaseConnection(Enum):\n    TCP = \'tcp\'\n    UDP = \'udp\'\n\nclass DatabaseConfig(BaseConfig):\n    ...\n    connection_type: DatabaseConnection\n```\n\n## Switching configuration based on Environment\nA very common situation is that configurations must be changed based on the execution environment (dev, test, prod). This can be accomplished \nby registering a folder with one .yaml file per environment and seting the `ENV` environment variable to the value you need. An example could look \nlike this:  \n\nLet\'s assume we have three environments (dev, test, prod) and one configuration file per environment in the following folder structure:\n```\nproject\n│\n└───config\n│   │   my_prod_config.yaml\n│   │   my_test_config.yaml\n│   │   my_dev_config.yaml\n│   \n└───src\n│   │   app.py\n│   │   my_config.py\n```\nThe definition of `my_config.py` is equivalent to the one used in the basic introduction section and `app.py` uses our configuration the \nfollowing way:\n```python\n# we register the folder where ConfME can find the configuration files\nMyConfig.register_folder(Path(__file__).parent / \'../config\')\n...\n\n# we access the instance of the corresponding configuration file anywhere in our project. \nmy_config = MyConfig.get()\nprint(f\'Using database connection {my_config.database.host} \'\n      f\'on port {my_config.database.port}\')\n```\nIf now one of the following environment variables (precedence in descending order): `[\'env\', \'environment\', \'environ\', \'stage\']` is \nset e.g. `export ENV=prod` it will load the configuration file with `prod` in its name.\n\n## Parameter overwrite\nIn addition to loading configuration parameters from the configuration file, they can be passed/overwritten from the command line or environment variables. Thereby, the following precedences apply (lower number means higher precedence):\n1. **Command Line Arguments**: Check if parameter is set as command line argument. If not go one line done...\n2. **Environment Variables**: Check if parameter is set as environment variable. If not go one line done...\n3. **Configuration File**: If value was not found in one of the previous sources, it will check in the configuration file.\n\n### Overwrite Parameters from Command Line\nEspecially in the Data Science and Machine Learning area it is useful to pass certain parameters for experimental purposes as command line arguments. Therefore, all properties defined in the configuration classes are automatically offered as command line arguments in the following format:\n\n**my_program.py:**\n\n```python\nfrom confme import BaseConfig\n\nclass DatabaseConfig(BaseConfig):\n    host: str\n    port: int\n    user: str\n\nclass MyConfig(BaseConfig):\n    name: int\n    database: DatabaseConfig\n\nconfig = MyConfig.load(\'test.yaml\')\n```\n\nWhen you now start your program from the command line with the ```++help``` argument, you get the full list of all configuration options. **CAVEAT!** In order to not interfere with other cli tools, the prefix - was changed to +:\n```shell\n$ python my_program.py --help\nusage: my_program.py [+h] [++name NAME] [++database.host DATABASE.HOST] [++database.port DATABASE.PORT] [++database.user DATABASE.USER]\n\noptional arguments:\n  +h, ++help            show this help message and exit\n\nConfiguration Parameters:\n  With the parameters specified bellow, the configuration values from the config file can be overwritten.\n\n  ++name NAME\n  ++database.host DATABASE.HOST\n  ++database.port DATABASE.PORT\n  ++database.user DATABASE.USER\n```\n\n### Overwrite Parameters with Environment Variables\nLikewise to overwriting parameters from the commandline you can also overwrite by passing environment variables. Therefore, simply set the environment variable in the same format as it would be passed as command line arguments and run your application:\n```shell\n$ export database.host=localhost\n$ python my_programm.py\n```\n\n## LICENSE\n\nConfMe is released under the [MIT](LICENSE) license.\n\n',
    'author': 'Iwan Silvan Bolzern',
    'author_email': 'iwan.bolzern@roche.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/iwanbolzern/confme',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from os import path
from typing import Dict

from confme.source_backend.backend_yaml import YamlFileParser

FILE_PARSER = [
    YamlFileParser()
]


def parse_file(file_path: str) -> Dict:
    """Parses the given file with the right file parser based on the filename ending of the
    given file_path
    :param file_path: path to the file
    :return: Dict with content of the file
    """
    ending = path.splitext(file_path)[-1]
    applicable_file_parsers = [p for p in FILE_PARSER if ending in p.get_endings()]
    if len(applicable_file_parsers) < 0:
        raise Exception(f'File Ending {ending} not known')
    if len(applicable_file_parsers) > 1:
        raise Exception(f'More than one parser registered for this file ending... 🧐')

    with open(file_path, 'r') as file:
        return applicable_file_parsers[0].parse(file)



import os

from pydantic.main import ModelMetaclass

from confme.utils.dict_util import flatten, InfiniteDict
from confme.utils.typing import get_schema


def env_overwrite(config_cls: ModelMetaclass):
    # extract possible parameters
    config_dict = get_schema(config_cls)
    parameters, _ = flatten(config_dict)

    # make env variables case insensitive
    keys, values = zip(*os.environ.items())
    keys = [k.casefold() for k in keys]

    # find passed arguments and fill it into the dict structure
    infinite_dict = InfiniteDict()
    for p in parameters:
        if p.casefold() in keys:
            i = keys.index(p.casefold())
            infinite_dict.expand(p.split('.'), values[i])

    return infinite_dict
